
import importlib.resources
from typing import Any
from src import paths
import json, logging, os

def restore_default_settings():
    """
    Restores current settings file to defaults.
    """
    # Ensure config parent directory exists
    if not os.path.exists(paths.CONFIG_DIR):
        os.makedirs(paths.CONFIG_DIR, exist_ok=True)

    # Restore default settings file
    with importlib.resources.open_text('src.resources', 'default_settings.json') as f:
        defaults = f.read()

    with open(paths.SETTINGS_FILE_PATH, "w") as f:
        f.write(defaults)

def load_settings_file() -> dict[str, Any]:
    """
    Read and parse the settings file as JSON.
    """
    try:
        with open(paths.SETTINGS_FILE_PATH, "r") as f:
            return json.load(f)
    except json.JSONDecodeError:
        logging.log(logging.ERROR, "JSON decode error while reading parsing settings file. Check if it contains valid JSON and fix invalid syntax.")
        exit()
    except FileNotFoundError:
        logging.log(logging.INFO, "Settings file doesn't exist yet! Creating it...")
        restore_default_settings()
        return load_settings_file()

def get_setting(setting_key: str) -> Any:
    """
    Get a setting from the settings file. Exits program if setting couldn't be found.
    """
    settings = load_settings_file()

    try:
        return settings[setting_key]
    except KeyError:
        logging.log(logging.ERROR, "Couldn't find setting with key: "+setting_key)
        exit()
    
def set_setting(setting_key: str, new_value: Any):
    """
    Set the value of a setting.
    """
    settings = load_settings_file()

    try:
        # NOTE: check if type is the same maybe?
        settings[setting_key] = new_value
    except KeyError:
        logging.log(logging.WARN, "Tried setting value of setting key that doesn't exist in the settings file. Settings not changed.")
        return
    
    with open(paths.SETTINGS_FILE_PATH, "w") as f:
        json.dump(settings, f)


from src import custom_logging, arguments, paths, tle
import logging, os, shutil
import importlib_resources # backport for pre python 3.12 compatibility

def main():
    # Set up logging
    custom_logging.set_up_logging()

    # Ensure necessary directories exist
    if not os.path.exists(paths.CONFIG_DIR):
        os.makedirs(paths.CONFIG_DIR, exist_ok=True)

    if not os.path.exists(paths.DATA_DIR):
        os.makedirs(paths.DATA_DIR, exist_ok=True)

    if not os.path.exists(paths.TLE_DIRECTORY_PATH):
        os.makedirs(paths.TLE_DIRECTORY_PATH, exist_ok=True)

    if not os.path.exists(paths.TRANSPONDERS_DIRECTORY_PATH):
        os.makedirs(paths.TRANSPONDERS_DIRECTORY_PATH, exist_ok=True)

    resources_files = importlib_resources.files().joinpath("resources")
    if not os.path.exists(paths.ROTOR_CONFIG_DIRECTORY_PATH):
        os.makedirs(paths.ROTOR_CONFIG_DIRECTORY_PATH, exist_ok=True)

        # Copy default config files
        with importlib_resources.as_file(resources_files.joinpath("rotors_default")) as dir:
            shutil.copytree(dir, paths.ROTOR_CONFIG_DIRECTORY_PATH, dirs_exist_ok=True)

    if not os.path.exists(paths.RADIO_CONFIG_DIRECTORY_PATH):
        os.makedirs(paths.RADIO_CONFIG_DIRECTORY_PATH, exist_ok=True)

        # Copy default config files
        with importlib_resources.as_file(resources_files.joinpath("radios_default")) as dir:
            shutil.copytree(dir, paths.RADIO_CONFIG_DIRECTORY_PATH, dirs_exist_ok=True)

    # Ensure necessary files exist
    if not os.path.exists(paths.SOURCES_PATH):
        with open(paths.SOURCES_PATH, "w"): pass
        tle.add_source("https://celestrak.org/NORAD/elements/gp.php?GROUP=amateur&FORMAT=json")
        tle.add_source("https://celestrak.org/NORAD/elements/gp.php?GROUP=weather&FORMAT=json")

    if not os.path.exists(paths.LAST_TLE_UPDATE_PATH):
        with open(paths.LAST_TLE_UPDATE_PATH, "w") as f:
            f.write("0")

    # Settings file is created when load_settings_file is called, so no need to do it here.

    # Check for TLE age
    TLE_age_human_readable = tle.get_TLE_age_human_readable()
    if TLE_age_human_readable == "never":
        logging.log(logging.WARN, "TLEs have never been updated. Update using `satgs update tles` when possible.")
    else:
        if tle.check_TLEs_outdated():
            logging.log(logging.WARN, f"TLEs are {TLE_age_human_readable} old. Update when possible.")
        else:
            logging.log(logging.INFO, f"TLEs are {TLE_age_human_readable} old.")

    # Set up argument parsing and parse args
    arguments.set_up_argparse()



from src import paths, util
from typing import Tuple
import os, requests, logging, json

SATNOGS_TRANSITTERS_API_URL = "https://db.satnogs.org/api/transmitters/"

TRANSPONDER_TYPES = {"Transponder": "T", "Transceiver": "R", "Transmitter": "B"}

def download_transponders():
    """
    Download the newest version of the transmitters.json file.
    """

    # Try to download data
    try:
        request = requests.get(SATNOGS_TRANSITTERS_API_URL, params={"format": "json"})
    except Exception as e:
        logging.log(logging.ERROR, "Failed to download transponder data.")
        logging.log(logging.ERROR, e)
        exit()

    # Check status code
    if request.status_code != 200:
        logging.log(logging.ERROR, f"Failed to download transponder data. API returned status code {request.status_code}.")
        exit()

    # Try to parse JSON data
    try:
        json_data = json.loads(request.text)
    except json.JSONDecodeError:
        logging.log(logging.ERROR, "Failed to download transponder data. API returned invalid JSON.")
        exit()
    
    # Sort transponders to a per NORAD ID dict
    transponders = {}
    for trsp in json_data:
        try: # If satellite already exists in formatted transponder list, just add the current transponder to the dict indexed by its uuid
            transponders[trsp["norad_cat_id"]][trsp["uuid"]] = trsp
        except KeyError: # If it doesn't exist yet, create the satellites entry and set it to a dict containing the current transponder indexed by it's uuid
            transponders[trsp["norad_cat_id"]] = {trsp["uuid"]: trsp}

    # Save to dict elements to files
    for NORAD_ID, trsp in transponders.items():
        with open(os.path.join(paths.TRANSPONDERS_DIRECTORY_PATH, str(NORAD_ID)+".json"), "w") as f:
            json.dump(trsp, f)

def get_transponder_frequencies(NORAD_ID: str, transponder_UUID: str) -> Tuple[int, int | None, int, int | None, bool]:
    """
    Get the uplink and downlink frequencies of a transponder by the satellite NORAD ID and the transponder UUID.
    This function returns a tuple with four elements.
    
    The firt element is the lower downlink frequency.
    The second element is the upper downlink frequency. If the downlink is not a range of frequencies, the first value should be used and this one will be None.
    The third and fourth elements are the same but for the uplink.
    """

    # Try to open file and parse json
    try:
        with open(os.path.join(paths.TRANSPONDERS_DIRECTORY_PATH, NORAD_ID+".json"), "r") as f:
            trsp = json.load(f)[transponder_UUID]
    except FileNotFoundError:
        logging.log(logging.ERROR, f"Error while loading transponder file for NORAD ID {NORAD_ID}. File does not exist.")
        exit()
    except json.JSONDecodeError:
        logging.log(logging.ERROR, f"Error while loading transponder file for NORAD ID {NORAD_ID}. File contains invalid JSON.")
        exit()

    return (trsp["downlink_low"], trsp["downlink_high"], trsp["uplink_low"], trsp["uplink_high"], trsp["invert"])

def user_transponder_selection(NORAD_ID: str) -> str:
    """
    Prompt the user to select a transponder out of all transponder available on the satellite. Returns the selected transponders UUID.
    """

    # Load transponder data for satellite
    with open(os.path.join(paths.TRANSPONDERS_DIRECTORY_PATH, NORAD_ID+".json"), "r") as f:
        transponders = json.load(f)

    # Find longest option for spacing (description doesn't matter as it's at the end)
    longest_num = len(str(len(transponders)+1))
    longest_mode = 0
    for _, trsp in transponders.items():
        mode_len = len(trsp["mode"])

        if mode_len > longest_mode:
            longest_mode = mode_len

    # Print options to user
    for i, (_, trsp) in enumerate(transponders.items()):
        num = str(i+1)
        mode = trsp["mode"]
        desc = trsp["description"]
        
        trsp_type = trsp["type"]
        trsp_type = TRANSPONDER_TYPES[trsp_type]
        downlink_freq = trsp["downlink_low"]
        downlink_freq = util.get_frequency_band_letter(downlink_freq) if downlink_freq else " " # if its none use a space
        uplink_freq = trsp["uplink_low"]
        uplink_freq = util.get_frequency_band_letter(uplink_freq) if uplink_freq else " " # if its none use a space
        freq = trsp_type + downlink_freq + uplink_freq

        num = " "*(longest_num-len(num))+num
        mode = mode+" "*(longest_mode-len(mode)) # pad with spaces

        logging.log(logging.INFO, f"{num}. {freq} {mode}  /  {desc}")

    # Get input
    choice = util.decorated_input()
    if choice.lower() == "help": # show help menu
        print()
        logging.log(logging.INFO ,"Transponder selection help menu")
        logging.log(logging.INFO ,"I. TTT MMMMM* / DDDDDDDDDD*")
        logging.log(logging.INFO ,"I: Option index")
        logging.log(logging.INFO ,"T: Transponder info")
        logging.log(logging.INFO ,"T: 1. character is either <T>ransponder, <R>epeater or <B>eacon")
        logging.log(logging.INFO ,"T: 2. character is downlink frequency band")
        logging.log(logging.INFO ,"T: 3. character is uplink frequency band")
        logging.log(logging.INFO ,"T: 3. frequency band letter will be 'H' if frequency if below VHF or 'O' if its above Ka band")
        logging.log(logging.INFO ,"M: Mode")
        logging.log(logging.INFO ,"D: Description")
        logging.log(logging.INFO ,"Press enter to return to selection")
        input()
        return user_transponder_selection(NORAD_ID)
    else: # return selection
        try:
            return list(transponders.items())[int(choice)-1][0]
        except Exception:
            logging.log(logging.ERROR, "Invalid choice!")
            exit()



from src import tle, paths
import os, datetime, re, logging, shutil, socket

last_port = 56000

COSPAR_ID_REGEX = re.compile(r'^[0-9]{4}-[0-9]{3}[A-Z]{1,3}$')

FREQUENCY_BAND_LETTERS = [
    (0, 30e6, "H"),          # HF and below
    (30e6, 300e6, "V"),      # VHF
    (300e6, 1e9, "U"),       # UHF
    (1e9, 2e9, "L"),         # L-Band
    (2e9, 4e9, "S"),         # S-Band
    (4e9, 8e9, "C"),         # C-Band
    (8e9, 12e9, "X"),        # X-Band
    (12e9, 40e9, "K")       # K-Band (includes Ku and Ka)
    # Anything else will be O (other)
]

def is_poetry():
    """A function to check if program is being run in poetry environment"""
    venv = os.environ.get("VIRTUAL_ENV")
    if not venv:
        return False
    name = os.path.basename(venv)
    if "poetry" in os.path.dirname(venv) or name.startswith(os.path.basename(os.getcwd()) + "-"):
        return True
    return os.environ.get("POETRY_ACTIVE") == "1"

def decorated_input() -> str:
    """A function to get input from the user while maintaining the logging style"""

    time = datetime.datetime.now().strftime("%H:%M:%S")
    decorator_string = f"[{time}] (I) >> "

    return input(decorator_string)

def _check_port_used(port: int) -> bool:
    """A function to check if a port is in use"""

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0
    
def get_frequency_band_letter(frequency: int) -> str:
    """A function to get a frequency band letter by the frequency in herz"""
    for lower, upper, letter in FREQUENCY_BAND_LETTERS:
        if frequency >= lower and frequency <= upper:
            return letter
    return "O"
    
def get_unused_port(purpose: str = "N/A") -> int:
    """
    Get an unused port. Optionally provide the purpose for the port for logging.
    """
    global last_port
    
    while True:
        last_port += 1

        if _check_port_used(last_port):
            logging.log(logging.DEBUG, f"Port {last_port} is in use")
        else:
            logging.log(logging.DEBUG, f"Using port {last_port} for {purpose}")
            return last_port

def satellite_norad_from_input(input: str) -> str:
    """
    Get a satellite NORAD ID by either one of these input opions:
    1. Just the NORAD ID
    2. The satellites name (input required if multiple matches)
    3. COSPAR ID

    Which of these was provided will be detected automatically.
    """

    input = input.strip()

    if input.isdigit(): # Check if input is a NORAD ID
        return input
    
    sat_IDs = tle.load_tle_data() # If it's not a NORAD ID, we have to load some extra data
    if COSPAR_ID_REGEX.match(input): # Check if input is a COSPAR ID
        index = [x for x, y in enumerate(sat_IDs) if y[1] == input]
        if index == []:
            logging.log(logging.ERROR, f"Can't find satellite with COSPAR ID '{input}' in local TLEs.")
            exit()
        sat_ID = sat_IDs[index[0]]
        
        return sat_ID[0]
    else: # Otherwise it's probably a name
        prepared_input = input.lower().replace("-", " ")
        search_hits = [t for t in sat_IDs if prepared_input in t[2].lower().replace("-", " ")]
        if search_hits == []:
            logging.log(logging.ERROR, f"Can't find satellite with name '{input}' in local TLEs.")
            exit()
        
        if len(search_hits) == 1:
            return search_hits[0][0]
        else:
            logging.log(logging.INFO, f"Found multiple hits while searching for '{input}'. Select the index of the satellite you wish to pick.")
            for i, satellite in enumerate(search_hits):
                logging.log(logging.INFO, f"{i+1}. {satellite[0]} \\ {satellite[2]}")
            choice = decorated_input()
            try:
                return search_hits[int(choice)-1][0]
            except (TypeError, ValueError, IndexError):
                logging.log(logging.ERROR, "Invalid choice!")
                exit()

def clean_all_data():
    """
    A function to delete all data files ever created by satgs.
    """

    logging.log(logging.INFO, "Deleting data..")
    shutil.rmtree(paths.CONFIG_DIR)
    shutil.rmtree(paths.DATA_DIR)
    logging.log(logging.INFO, "Done!")



import platformdirs, os

# Definitions for paths of all program files
CONFIG_DIR = platformdirs.user_config_dir("satgs")
DATA_DIR = platformdirs.user_data_dir("satgs")

TLE_DIRECTORY_PATH = os.path.join(DATA_DIR, "tle/")
TRANSPONDERS_DIRECTORY_PATH = os.path.join(DATA_DIR, "transponders/")
SOURCES_PATH = os.path.join(CONFIG_DIR, "sources.txt")
LAST_TLE_UPDATE_PATH = os.path.join(DATA_DIR, "last_tle_update.txt")

ROTOR_CONFIG_DIRECTORY_PATH = os.path.join(CONFIG_DIR, "rotors/")
RADIO_CONFIG_DIRECTORY_PATH = os.path.join(CONFIG_DIR, "radios/")

SETTINGS_FILE_PATH = os.path.join(CONFIG_DIR, "settings.json")



from src import paths, settings
from skyfield.api import EarthSatellite
from skyfield.timelib import Timescale
from typing import List, Tuple
import logging, json, requests, os, datetime

TLE_OUTDATED_SECONDS = int(settings.get_setting("tles_outdated_seconds")) # Hours until TLEs will be considered out of date in seconds

EXPECTED_TLE_JSON_KEYS = ['OBJECT_NAME', 'OBJECT_ID', 'EPOCH', 'MEAN_MOTION', 'ECCENTRICITY', 'INCLINATION', 'RA_OF_ASC_NODE', 'ARG_OF_PERICENTER', 'MEAN_ANOMALY', 'EPHEMERIS_TYPE', 'CLASSIFICATION_TYPE', 'NORAD_CAT_ID', 'ELEMENT_SET_NO', 'REV_AT_EPOCH', 'BSTAR', 'MEAN_MOTION_DOT', 'MEAN_MOTION_DDOT']

def check_source(source_url: str, print_failure_reason: bool = True) -> bool:
    """
    Checks if a given source URL provides valid json format TLE data. This can either be a single TLE, or an array of TLEs.
    """
    level = logging.DEBUG
    if print_failure_reason:
        level = logging.WARN
    
    logging.log(logging.DEBUG, "Checking TLE source: "+source_url)
    
    # Attempt to download the TLE file
    try:
        TLE_request = requests.get(source_url)
    except Exception as e:
        logging.log(level, "TLE source failed check: Failed to download data. "+str(e))
        return False
    
    # Check status code of data
    if TLE_request.status_code != 200:
        logging.log(level, f"TLE source failed check: Returned status code {str(TLE_request.status_code)}.")
        return False

    # Try parsing the data to check if it's valid json
    try:
        json_data = json.loads(TLE_request.text)
    except json.JSONDecodeError:
        logging.log(level, "TLE source failed check: JSON decode error.")
        return False
    
    # Check if required fields are present
    if type(json_data) is list: # Multiple TLEs
        for TLE in json_data:
            if EXPECTED_TLE_JSON_KEYS != list(TLE.keys()):
                logging.log(level, "TLE source failed check: Invalid JSON keys.")
                return False
    elif type(json_data) is dict: # Single TLE
        if EXPECTED_TLE_JSON_KEYS != list(json_data.keys()):
            logging.log(level, "TLE source failed check: Invalid JSON keys.")
            return False
    else:
        logging.log(level, "TLE source failed check: Invalid data type after parsing.")
        return False

    logging.log(logging.DEBUG, "TLE source checked successfully.")

    return True

def add_source(source_url: str):
    """
    Add a source URL to the list of TLE sources. The source must provide valid json TLEs.
    It can provide multiple TLEs in an array of TLEs.
    """

    # Check if source provides valid data
    if not check_source(source_url):
        logging.log(logging.WARN, "TLE source didn't provide valid data. Not adding source.")
        return

    # Ensure newline at the end of each line
    if not source_url.endswith("\n"):
        source_url += "\n"

    # Append to file
    with open(paths.SOURCES_PATH, "a") as f:
        f.write(source_url)
    
    logging.log(logging.DEBUG, f"Added TLE source {source_url} to sources file.")

def remove_source(source_url: str):
    """
    Remove a source URL from the list of TLE sources. The TLE provided by the source will only be removed on the next update.
    """

    # Read lines
    with open(paths.SOURCES_PATH, "r") as f:
        lines = f.readlines()

    # If source is in list of lines, remove it
    try:
        source_line_index = lines.index(source_url+"\n")
        del lines[source_line_index]
    except ValueError:
        logging.log(logging.WARN, f"Failed to remove source {source_url}. Source not found in sources file.")
        return

    # Write new sources to file
    with open(paths.SOURCES_PATH, "w") as f:
        f.writelines(lines)

def list_sources() -> List[str]:
    """
    Logs a list of all sources with indexes, and returns a list of the source URLs corrensponding to the logged indexes-1
    """
    # Read sources
    with open(paths.SOURCES_PATH, "r") as f:
        sources = f.readlines()

    # Prepare output and 
    logged_sources = []
    for source in sources:
        if source == "\n":
            continue
        source = source.strip()
        logging.log(logging.INFO, f"{len(logged_sources)+1}. {source}")
        logged_sources.append(source)

    return logged_sources

def _process_TLE(data: dict[str, str | int], source: str):
    """
    Internal function used by `download_TLEs` to save TLE data after it had been downloaded and parsed.
    """

    # Check if keys in data are the expected keys
    if EXPECTED_TLE_JSON_KEYS != list(data.keys()):
        logging.log(logging.WARN, f"Failed to download TLEs from source {source}. Source provided data with invalid keys. Skipping this source.")
        return

    # Save data to file
    NORAD_ID = data["NORAD_CAT_ID"]
    with open(os.path.join(paths.TLE_DIRECTORY_PATH, str(NORAD_ID)+".json"), "w") as f:
        json.dump(data, f)

def download_TLEs(log_progress: bool = True):
    """
    Download all TLEs from sources in sources file.
    """
    # Read sources file
    with open(paths.SOURCES_PATH, "r") as f:
        sources = f.readlines()

    # Count total sources that aren't new lines
    total_sources = 0
    for source in sources:
        if source != "\n":
            total_sources += 1

    # If log_progress is enabled, set level of progress logs to info instead of debug
    progress_log_level = logging.DEBUG
    if log_progress:
        progress_log_level = logging.INFO

    # Delete old TLE files
        for file in os.listdir(paths.TLE_DIRECTORY_PATH):
            full_path = os.path.join(paths.TLE_DIRECTORY_PATH, file)
            os.remove(full_path)

    # Process each source
    skipped = 0
    for i, source in enumerate(sources):
        if source == "\n": # Skip empty new lines
            skipped += 1
            continue

        source = source.strip()

        # Log progress
        progress_percent = ((i-skipped+1)/total_sources)*100
        progress_bar = f"[{'='*(round(progress_percent/10))}{' '*(10-round(progress_percent/10))}]"
        logging.log(progress_log_level, f"Updating TLEs... {progress_bar} ({i-skipped+1}/{total_sources})")
        
        # Try to download TLEs
        try:
            TLE_request = requests.get(source)
        except Exception as e:
            logging.log(logging.WARN, f"Failed to download TLEs from source {source}. {str(e)}. Skipping this source.")
            continue

        # Check status code
        if TLE_request.status_code != 200:
            logging.log(logging.WARN, f"Failed to download TLEs from source {source}. Source returned status code {TLE_request.status_code}. Skipping this source.")
            continue

        # Try to parse json TLE data
        try:
            TLE_json_data = json.loads(TLE_request.text)
        except json.JSONDecodeError:
            logging.log(logging.WARN, f"Failed to download TLEs from source {source}. Source provided invalid JSON. Skipping this source.")
            continue
        
        # Check if source provided a single or multiple TLEs
        if type(TLE_json_data) is list: # multiple
            for TLE in TLE_json_data:
                _process_TLE(TLE, source)
        elif type(TLE_json_data) is dict: # single
            _process_TLE(TLE_json_data, source)
        else:
            logging.log(logging.WARN, f"Failed to download TLEs from source {source}. Source provided data that caused an invalid data type after parsing. Skipping this source.")
            continue

    # Update last TLE update timestamp
    timestamp = int(datetime.datetime.now(datetime.timezone.utc).timestamp())
    with open(paths.LAST_TLE_UPDATE_PATH, "w") as f:
        f.write(str(timestamp))

def get_last_TLE_update() -> datetime.datetime:
    """
    Get the time of the last TLE update as a datetime object
    """

    with open(paths.LAST_TLE_UPDATE_PATH, "r") as f:
        ts = f.read()
    return datetime.datetime.fromtimestamp(int(ts), datetime.timezone.utc)

def get_TLE_age_human_readable() -> str:
    """
    Returns a string about how long ago the last TLE update was in a human readable format.
    Will return "never" if TLEs have never been updated.
    """
    last_update = get_last_TLE_update()

    # Check if TLEs have never been updated (assuming time travel isn't possible)
    if last_update.timestamp() == 0:
        return "never"

    delta = datetime.datetime.now(datetime.timezone.utc) - last_update
    seconds = delta.total_seconds()

     # Define units and how many seconds they equal
    intervals = (
        ('year', 365 * 24 * 3600),
        ('month', 30 * 24 * 3600),
        ('day', 24 * 3600),
        ('hour', 3600),
        ('minute', 60),
        ('second', 1),
    )

    for name, count in intervals:
        value = seconds // count
        if value:
            # Proper pluralization
            unit = name
            if value != 1:
                unit = name + "s"
            return f"{round(value)} {unit}"
        
    return "1 second"

def check_TLEs_outdated() -> bool:
    """
    Returns True if TLEs are older than 96 hours.
    """
    last_update = get_last_TLE_update()
    delta = datetime.datetime.now(datetime.timezone.utc) - last_update

    return delta.total_seconds() > int(TLE_OUTDATED_SECONDS)

def load_tle_data() -> List[Tuple[str, str, str]]:
    """
    Load NORAD IDs, COSPAR IDs and names for all available satellites and return them in a list of tuples (NORAD, COSPAR, name).
    """

    data = []
    tle_files = os.listdir(paths.TLE_DIRECTORY_PATH)
    for tle_file in tle_files:
        with open(os.path.join(paths.TLE_DIRECTORY_PATH, tle_file), "r") as f:
            tle_data = json.load(f)
        data.append((str(tle_data["NORAD_CAT_ID"]), tle_data["OBJECT_ID"], tle_data["OBJECT_NAME"]))
    
    return data

def load_tle(NORAD_ID: str, timescale: Timescale) -> EarthSatellite | None:
    """
    Load a satellite TLE by it's NORAD ID. Will return None if TLE can't be found in local files.
    If TLE is found, this function will return a skyfield `EarthSatellite` object.
    """

    # Try to open TLE file
    try:
        with open(os.path.join(paths.TLE_DIRECTORY_PATH, NORAD_ID+".json"), "r") as f:
            TLE_data = json.load(f)

        # Initialize and return EarthSatellite object.
        return EarthSatellite.from_omm(timescale, TLE_data)
    except FileNotFoundError:
        # If it doesn't exist, return none
        return None



from src import paths, util
from typing import Dict
import subprocess, os, json, logging, socket

RADIO_SDR_CONF_EXPECTED_KEYS = set(["rigctl_port"])
RADIO_RX_CONF_EXPECTED_KEYS = set(["usb_port", "rigctl_ID", "rigctl_port_overwrite", "serial_speed", "offset"])
RADIO_TX_CONF_EXPECTED_KEYS = RADIO_RX_CONF_EXPECTED_KEYS

def parse_radio_config(radio_config_name: str) -> Dict[str, Dict[str, str | int]]:
    """
    Parse a radio config file by its file name (excluding file exension).
    Returns all values specified in the README section for the radio config files in a dictionary.
    """
    
    # Try to read and parse JSON data
    file_path = os.path.join(paths.RADIO_CONFIG_DIRECTORY_PATH, radio_config_name+".json")
    try:
        with open(file_path, "r") as f:
            json_data = json.load(f)
    except json.JSONDecodeError as e:
        logging.log(logging.ERROR, "Failed parsing file radio config file '"+radio_config_name+".json'.")
        logging.log(logging.ERROR, e)
        exit()

    # Checks to see if data is valid
    if type(json_data) is not dict:
        logging.log(logging.ERROR, "Failed parsing file radio config file '"+radio_config_name+".json'. JSON data parsed to invalid data type.")
        exit()

    # Make sure that atleast one valid radio type is defined and that the defined ones have the expected keys.
    valid_radio_type_defined = False
    if "sdr" in json_data:
        if set(json_data["sdr"].keys()) != RADIO_SDR_CONF_EXPECTED_KEYS:
            logging.log(logging.ERROR, "Failed parsing file radio config file '"+radio_config_name+".json'. Invalid keys present in SDR section of config file.")
            exit()
        valid_radio_type_defined = True
    
    if "rx" in json_data:
        if set(json_data["rx"].keys()) != RADIO_RX_CONF_EXPECTED_KEYS:
            logging.log(logging.ERROR, "Failed parsing file radio config file '"+radio_config_name+".json'. Invalid keys present in RX section of config file.")
            exit()
        valid_radio_type_defined = True
    
    if "tx" in json_data:
        if set(json_data["tx"].keys()) != RADIO_TX_CONF_EXPECTED_KEYS:
            logging.log(logging.ERROR, "Failed parsing file radio config file '"+radio_config_name+".json'. Invalid keys present in TX section of config file.")
            exit()
        valid_radio_type_defined = True

    # TODO: transceiver

    if not valid_radio_type_defined:
        logging.log(logging.ERROR, "Failed parsing file radio config file '"+radio_config_name+".json'. Couldn't find any valid radio types defined.")

    return json_data

class Radio_Controller():
    def __init__(self, radio_config_name: str, downlink_frequency: int | None, uplink_frequency: int | None, rx_usb_overwrite: str | None, tx_usb_overwrite: str | None, trx_usb_overwrite: str | None, inverting: bool = False, lock: bool = True) -> None:
        """
        Initialize radio object. Must provide the name of the radio config file to be read (without the file extension).
        Optionally the downlink and uplink frequency of the satellite transponder, if it is inverting, USB port overwrites and wether to lock downlink and uplink together can be provided.
        """
        
        self.downlink_freq = downlink_frequency
        self.uplink_freq = uplink_frequency

        self.lock = lock
        self.inversion_multi = -1 if inverting else 1 # A multiplier to be applied to an offset if the transponder is inverting

        self.corrected_downlink = None
        self.corrected_uplink = None

        # Parse config
        radio_config = parse_radio_config(radio_config_name)

        if (("tx" in radio_config) or ("trx" in radio_config)) and uplink_frequency is None: # Warn user if a transmitter has been defined but no uplink freq was provided
            logging.log(logging.WARN, "Transmitter/Transceiver is defined in the configuration but no uplink frequency was provided. Transmitters will be ignored.")

        if (("sdr" in radio_config) or ("rx" in radio_config) or ("trx" in radio_config)) and downlink_frequency is None: # Warn user if a receiver has been defined but no downlink freq was provided
            logging.log(logging.WARN, "Receiver/Transceiver/SDR is defined in the configuration but no downlink frequency was provided. Receivers will be ignored.")

        if ("tx" in radio_config) and ("rx" in radio_config):
            if radio_config["tx"]["usb_port"] == radio_config["rx"]["usb_port"]:
                logging.log(logging.ERROR, "Defined transmitter and receiver can't be the same device! Try defining a transceiver instead.")
                exit()

        # Initialize SDR if present in config
        if "sdr" in radio_config:
            sdr_config = radio_config["sdr"]
            self.sdr_rigctld_port = int(sdr_config["rigctl_port"])

            # Try to connect to SDR rigctl
            try:
                logging.log(logging.DEBUG, "Opening socket to rigctl (SDR)")
                self.sdr_sock = socket.create_connection(("localhost", int(self.sdr_rigctld_port)), timeout=3)
            except Exception as e:
                logging.log(logging.ERROR, "Failed to open connection to SDR rigctl server. Skipping this radio.")
                logging.log(logging.ERROR, e)
                self.sdr_sock = None
        else:
            self.sdr_sock = None

        # Initialize receiver in config
        if "rx" in radio_config:
            rx_config = radio_config["rx"]
            self.rx_usb_port = rx_usb_overwrite if rx_usb_overwrite else rx_config["usb_port"]
            self.rx_rigctl_ID = rx_config["rigctl_ID"]
            self.rx_serial_speed = rx_config["serial_speed"]
            self.rx_offset = rx_config["offset"]
            self.rx_rigctld_port = util.get_unused_port("rigctld (receiver)") if rx_config["rigctl_port_overwrite"] == 0 else rx_config["rigctl_port_overwrite"]

            # Check if offset and speed are ints
            try:
                self.rx_serial_speed = int(self.rx_serial_speed)
            except ValueError:
                logging.log(logging.ERROR, f"Configured receiver serial speed '{self.rx_serial_speed}' is not a valid integer.")
                exit()
            try:
                self.rx_offset = int(self.rx_offset)
            except ValueError:
                logging.log(logging.ERROR, f"Configured receiver offset '{self.rx_offset}' is not a valid integer.")
                exit()

            # Attempt to start rigctld
            logging.log(logging.INFO, "Starting rigctld (receiver)")
            self.rx_rigctld = subprocess.Popen(
                ["rigctld", "-m", str(self.rx_rigctl_ID), "-r", str(self.rx_usb_port), "-t", str(self.rx_rigctld_port), "-s", str(self.rx_serial_speed)],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True
            )

            try:
                stdout, stderr = self.rx_rigctld.communicate(timeout=1)
                if self.rx_rigctld.returncode != 0:
                    logging.log(logging.ERROR, "Rigctld failed with error code "+str(self.rx_rigctld.returncode))
                    logging.log(logging.ERROR, "Error: "+str(stderr))
                    if stderr == "rig_open: error = IO error":
                        logging.log(logging.INFO, "Tip: Make sure you have the correct USB port selected." \
                                                  "You can overwrite the USB port in the config file using -r")
                    exit()
                if self.rx_rigctld is None:
                    logging.log(logging.ERROR, "Rigctld failed to start.")
            except subprocess.TimeoutExpired: # Rigctld is running
                pass

            # Open socket to rigctld
            try:
                logging.log(logging.DEBUG, "Opening socket to rigctl (receiver)")
                self.rx_sock = socket.create_connection(("localhost", int(self.rx_rigctld_port)), timeout=3)
            except Exception as e:
                logging.log(logging.ERROR, "Failed to open connection to receiver rigctl server. Skipping this radio.")
                logging.log(logging.ERROR, e)
                self.rx_sock = None
        else:
            self.rx_sock = None

        # Initialize transmitter in config (this is the same as the receiver part)
        if "tx" in radio_config:
            tx_config = radio_config["tx"]
            self.tx_usb_port = tx_usb_overwrite if tx_usb_overwrite else tx_config["usb_port"]
            self.tx_rigctl_ID = tx_config["rigctl_ID"]
            self.tx_serial_speed = tx_config["serial_speed"]
            self.tx_offset = tx_config["offset"]
            self.tx_rigctld_port = util.get_unused_port("rigctld (transmitter)") if tx_config["rigctl_port_overwrite"] == 0 else tx_config["rigctl_port_overwrite"]

            # Check if offset and speed are ints
            try:
                self.tx_serial_speed = int(self.tx_serial_speed)
            except ValueError:
                logging.log(logging.ERROR, f"Configured transmitter serial speed '{self.tx_serial_speed}' is not a valid integer.")
                exit()
            try:
                self.tx_offset = int(self.tx_offset)
            except ValueError:
                logging.log(logging.ERROR, f"Configured transmitter offset '{self.tx_offset}' is not a valid integer.")
                exit()

            # Attempt to start rigctld
            logging.log(logging.INFO, "Starting rigctld (transmitter)")
            self.tx_rigctld = subprocess.Popen(
                ["rigctld", "-m", str(self.tx_rigctl_ID), "-r", str(self.tx_usb_port), "-t", str(self.tx_rigctld_port), "-s", str(self.tx_serial_speed)],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True
            )

            try:
                stdout, stderr = self.tx_rigctld.communicate(timeout=1)
                if self.tx_rigctld.returncode != 0:
                    logging.log(logging.ERROR, "Rigctld failed with error code "+str(self.tx_rigctld.returncode))
                    logging.log(logging.ERROR, "Error: "+str(stderr))
                    if stderr == "rig_open: error = IO error":
                        logging.log(logging.INFO, "Tip: Make sure you have the correct USB port selected." \
                                                  "You can overwrite the USB port in the config file using -t")
                    exit()
                if self.tx_rigctld is None:
                    logging.log(logging.ERROR, "Rigctld failed to start.")
            except subprocess.TimeoutExpired: # Rigctld is running
                pass

            # Open socket to rigctld
            try:
                logging.log(logging.DEBUG, "Opening socket to rigctl (transmitter)")
                self.tx_sock = socket.create_connection(("localhost", int(self.tx_rigctld_port)), timeout=3)
            except Exception as e:
                logging.log(logging.ERROR, "Failed to open connection to transmitter rigctl server. Skipping this radio.")
                logging.log(logging.ERROR, e)
                self.tx_sock = None
        else:
            self.tx_sock = None

        # TODO: transceiver

        self.downlink_correction = 0 # Doppler correction factor in hz 
        self.uplink_correction = 0
        self.current_downlink_frequency = downlink_frequency if downlink_frequency is not None else 0 # Frequency that the doppler correction will be applied to
        self.current_uplink_frequency = uplink_frequency if uplink_frequency is not None else 0

    def _send_rigctl_command(self, sock: socket.socket, cmd: str):
        """
        Send a command to a specified rigctl(d) and return response lines (without newlines).
        """
        logging.log(logging.DEBUG, f"Sending rigctl command '{cmd}'")

        sock.sendall((cmd + '\n').encode('ascii'))
        response = sock.recv(4096).decode('ascii') # if this line fails its probably a config error
    
        # multiple lines, strip trailing newline
        return [line.strip() for line in response.splitlines()]

    def _set_frequency(self, sock: socket.socket, freq: int):
        """
        Send a rigctl(d) command to a specified socket to change the rig frequency. Frequency must be in herz.
        """

        self._send_rigctl_command(sock, f"F {freq}")

    def _read_frequency(self, sock: socket.socket, direction: str) -> int:
        """
        Send a rigctl(d) command to a specified socket to read the current rig frequency.
        Direction must be provided (either "uplink" or "downlink") to remove doppler correction from the frequency reading.
        Returns frequency in herz.
        """

        freq = int(self._send_rigctl_command(sock, "f")[0])
        if direction == "uplink":
            freq -= self.uplink_correction
                
            return round(freq)
        elif direction == "downlink":
            freq -= self.downlink_correction

            return round(freq)
        else:
            logging.log(logging.ERROR, "Invalid direction specified for read frequency command. This should never happen.")
            exit()

    def update_lock(self):
        """
        Synchronise the frequencies of uplink and downlink devices. The `update` function must be called to apply these updated frequencies.
        """

        # Meassure frequencies of all downlink radios
        down_freqs = {}
        if self.sdr_sock:
            down_freqs["sdr"] = self._read_frequency(self.sdr_sock, "downlink")
        
        if self.rx_sock: 
            down_freqs["rx"] = self._read_frequency(self.rx_sock, "downlink")

        # Meassure frequencies of all uplink radios
        up_freqs = {}
        if self.tx_sock:
            up_freqs["tx"] = self._read_frequency(self.tx_sock, "uplink")

        # Check which downlink device has the greatest frequency offset from the current frequency
        down_offset = 0
        down_offset_abs = -1

        for _, frequency in down_freqs.items():
            offset = frequency - self.current_downlink_frequency
            offset_abs = abs(offset)
            if offset_abs > down_offset_abs:
                down_offset_abs = offset_abs
                down_offset = offset
        
        # Check which uplink device has the greatest frequency offset from the current frequency
        up_offset = 0
        up_offset_abs = -1

        for _, frequency in up_freqs.items():
            offset = frequency - self.current_uplink_frequency
            offset_abs = abs(offset)
            if offset_abs > up_offset_abs:
                up_offset_abs = offset_abs
                up_offset = offset
        
        # Check if it's necessary to update the other devices
        if (up_offset_abs + down_offset_abs) < 4:
            return

        # Check which radio has the biggest offset to know what offset should be applied to all other radios and set offsets
        downlink_offset = 0
        uplink_offset = 0
        if down_offset_abs < up_offset_abs: # uplink is the guide offset
            uplink_offset = up_offset
            if self.lock:
                downlink_offset = up_offset * self.inversion_multi
        else: # downlink is the guide offset
            downlink_offset = down_offset
            if self.lock:
                uplink_offset = down_offset * self.inversion_multi

        # Apply offsets
        if self.downlink_freq:
            self.current_downlink_frequency += downlink_offset
        if self.uplink_freq:
            self.current_uplink_frequency += uplink_offset

    def update(self, range_rate: float):
        """
        Update all defined transmitters/receivers with the satellites range rate specified in km/s.
        """

        # Handle downlink
        if self.downlink_freq:
            # Calulate corrected frequency
            self.downlink_correction = -(range_rate / 299792.458) * self.current_downlink_frequency # type: ignore
            self.corrected_downlink = round(self.downlink_correction + self.current_downlink_frequency)

            # Update downlink listeners
            if self.sdr_sock:
                self._set_frequency(self.sdr_sock, self.corrected_downlink) # type: ignore

            if self.rx_sock:
                self._set_frequency(self.rx_sock, self.corrected_downlink+self.rx_offset) # type: ignore

        # Handle uplink
        if self.uplink_freq:
            # Calulate corrected frequency
            self.uplink_correction = -(range_rate / 299792.458) * self.uplink_freq # type: ignore
            self.corrected_uplink = round(self.uplink_correction + self.current_uplink_frequency)

            # Update uplink listeners
            if self.tx_sock:
                self._set_frequency(self.tx_sock, round(self.corrected_uplink)+self.tx_offset) # type: ignore

    def close(self):
        """Close all sockets and terminate rigctl instances"""
        logging.log(logging.DEBUG, "Closing radio controller")

        if self.sdr_sock:
            self.sdr_sock.close()
        
        if self.rx_sock:
            self.rx_sock.close()
            self.rx_rigctld.terminate()

        if self.tx_sock:
            self.tx_sock.close()
            self.tx_rigctld.terminate()



from src import tle, util, tracking, settings, paths, transponders, test
import argparse, logging

def set_debug():
    """A function to set the logging level to debug"""
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    for handler in root.handlers: # theoretically there should only be one handler but who knows
        handler.setLevel(logging.DEBUG)

def show_version(_args):
    """A function to log the currently running version and exit"""

    import subprocess
    import importlib.metadata

    latest_commit_hash = ""
    git_branch = ""

    try: # No garantee that git is installed
        git_branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).strip().decode('utf-8')
        latest_commit_hash = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()
    except Exception as e:
        logging.log(logging.DEBUG, "Error while checking git branch & lastest commit hash.")
        logging.log(logging.DEBUG, e)

    print()

    if latest_commit_hash == "":
        logging.log(logging.INFO, f"You are running satgs v{importlib.metadata.version('satgs')}")
    else:
        logging.log(logging.INFO, f"You are running satgs v{importlib.metadata.version('satgs')} ({git_branch}/{latest_commit_hash})")

    exit()

def no_args_message(_args):
    logging.log(logging.ERROR, "Please provide a subcommand to run. Run `satgs --help` for help.")
    

# update subcommand functions
def update_all(_args):
    logging.log(logging.INFO, "Updating TLEs..")
    tle.download_TLEs()
    logging.log(logging.INFO, "Updating transponders")
    transponders.download_transponders()
    logging.log(logging.INFO, "Done!")
    exit()

def update_TLEs(_args):
    tle.download_TLEs()
    logging.log(logging.INFO, "Done!")
    exit()

def update_transponders(_args):
    transponders.download_transponders()
    logging.log(logging.INFO, "Done!")
    exit()

# sources subcommand functions
def add_source(_args):
    logging.log(logging.INFO, "Enter the URL to the source your would like to add.")
    source_url = util.decorated_input()
    tle.add_source(source_url)
    exit()

def list_sources(_args):
    logging.log(logging.INFO, "Listing sources...")
    tle.list_sources()
    exit()

def remove_source(_args):
    logging.log(logging.INFO, "Listing sources...")
    sources = tle.list_sources()
    logging.log(logging.INFO, "Enter the index of the source you'd like to remove.")
    try:
        remove_index = int(util.decorated_input())-1
        tle.remove_source(sources[remove_index])
    except (TypeError, ValueError, IndexError):
        logging.log(logging.ERROR, "Invalid choice!")

# tracking subcommand
def track(args):
    # just a wrapper to accept the arguments
    rotor_mode_overwrite = None
    if args.rotor_normal:
        rotor_mode_overwrite = 1
    elif args.rotor_inverted:
        rotor_mode_overwrite = 2

    tracking.track(util.satellite_norad_from_input(args.satellite), args.rotor, args.radio, args.rotor_usb, args.rx_usb, args.tx_usb, args.trx_usb, not args.unlock, rotor_mode_overwrite)

# testing subcommands
def test_rotor(args):
    rotor_mode_overwrite = None
    if args.rotor_normal:
        rotor_mode_overwrite = 1
    elif args.rotor_inverted:
        rotor_mode_overwrite = 2

    test.rotor_test(args.rotor, args.rotor_usb, rotor_mode_overwrite)

def test_rotor_full(args):
    rotor_mode_overwrite = None
    if args.rotor_normal:
        rotor_mode_overwrite = 1
    elif args.rotor_inverted:
        rotor_mode_overwrite = 2

    test.rotor_test_full(args.rotor, args.rotor_usb, rotor_mode_overwrite)

def test_rotor_home(args):
    rotor_mode_overwrite = None
    if args.rotor_normal:
        rotor_mode_overwrite = 1
    elif args.rotor_inverted:
        rotor_mode_overwrite = 2

    test.rotor_home(args.rotor, args.rotor_usb, rotor_mode_overwrite)

def test_radio(args):
    test.test_radio(args.radio, args.downlink, args.uplink, args.rx_usb, args.tx_usb, args.trx_usb)

# settings subcommands
def list_settings(_args):
    logging.log(logging.INFO, "Listing settings...")
    settings_data = settings.load_settings_file()

    for setting, value in settings_data.items():
        logging.log(logging.INFO, setting+": "+str(value))
    exit()

def modify_setting(args):
    logging.log(logging.INFO, f"Setting value of setting '{args.setting_key}' to '{args.new_setting_value}'")
    settings.set_setting(args.setting_key, args.new_setting_value)
    exit()

def get_settings_path(_args):
    logging.log(logging.INFO, "Settings path: "+paths.CONFIG_DIR)

def clean_data(_args):
    logging.log(logging.INFO, "Are you sure that you want to delete all settings (rotor config, radio config, sources, settings)? (N/y)")
    choice = util.decorated_input()
    if choice.lower().strip() == "y":
        util.clean_all_data()

def set_up_argparse():
    # Set up base parser
    parser = argparse.ArgumentParser(prog="satgs", add_help=True)
    parser.add_argument("--debug", action="store_true",
                        help="set logging level to debug")
    parser.add_argument("--version", action="store_true",       # Alias for `$ satgs version`
                        help="show currently installed version")

    # Set up subcommands
    sub_parsers = parser.add_subparsers()

    # version subcommand
    parser_version = sub_parsers.add_parser("version", help="Show version")
    parser_version.set_defaults(func=show_version)

    # update subcommands
    parser_update = sub_parsers.add_parser("update", help="Update data like TLEs. Can be called without subcommand to update everything")
    update_sub = parser_update.add_subparsers(required=False)

    parser_update_tle = update_sub.add_parser("tles", help="Update all TLEs")
    parser_update_tle.set_defaults(func=update_TLEs)

    parser_update_transponders = update_sub.add_parser("transponders", help="Update transponders file")
    parser_update_transponders.set_defaults(func=update_transponders)

    parser_update.set_defaults(func=update_all)

    # sources subcommands
    parser_sources = sub_parsers.add_parser("sources", help="Manage TLE sources")
    sources_sub = parser_sources.add_subparsers(required=True)

    parser_sources_add = sources_sub.add_parser("add", help="Add a TLE source")
    parser_sources_add.set_defaults(func=add_source)

    parser_sources_list = sources_sub.add_parser("list", help="List all current TLE sources")
    parser_sources_list.set_defaults(func=list_sources)

    parser_sources_remove = sources_sub.add_parser("remove", help="Remove a TLE source")
    parser_sources_remove.set_defaults(func=remove_source)

    # common parser for arguments shared between tracking and testing subcommands (argparse is very weird)
    parser_control_common = argparse.ArgumentParser(add_help=False)
    parser_control_common.add_argument("--rotor", type=str, choices=tracking.list_rotors(), dest="rotor",
                              help="The name of a rotor config file (without file extension)")
    parser_control_common.add_argument("--radio", type=str, choices=tracking.list_radios(),
                              help="The name of a radio config file (without file extension)")
    parser_control_common.add_argument("-o", "--rotor_usb", type=str,
                              help="USB port that the rotor is connected to (optional)")
    parser_control_common.add_argument("-n", "--rotor_normal", action="store_true",
                              help="Overwrite the rotor config to use rotor control mode 1")
    parser_control_common.add_argument("-i", "--rotor_inverted", action="store_true",
                              help="Overwrite the rotor config to use rotor control mode 2")
    parser_control_common.add_argument("-r", "--rx_usb", type=str,
                              help="USB port that the receiver is connected to (optional)")
    parser_control_common.add_argument("-t", "--tx_usb", type=str,
                              help="USB port that the transmitter is connected to (optional)")
    parser_control_common.add_argument("-x", "--trx_usb", type=str,
                              help="USB port that the transceiver is connected to (optional)")

    # tracking subcommand
    parser_track = sub_parsers.add_parser("track", help="Track a satellite", parents=[parser_control_common])
    parser_track.add_argument("satellite", help="Either the NORAD ID, COSPAR ID or name of the satellite to be tracked." \
                                                  "TLE must be avaliable in local database.")
    parser_track.add_argument("-u", "--unlock", action="store_true",
                              help="Don't lock uplink and downlink together for satellites with a frequency range.")
    parser_track.set_defaults(func=track)

    # testing subcommands
    parser_test = sub_parsers.add_parser("test", help="Test a rotor or radio", parents=[parser_control_common])
    test_sub = parser_test.add_subparsers(required=True)

    parser_test_rotor = test_sub.add_parser("rotor", help="Test a rotor by moving it a bit or select another rotor test", parents=[parser_control_common])
    parser_test_rotor.set_defaults(func=test_rotor)
    test_rotor_sub = parser_test_rotor.add_subparsers(required=False)

    parser_test_rotor_home = test_rotor_sub.add_parser("home", help="Home a rotor", parents=[parser_control_common])
    parser_test_rotor_home.set_defaults(func=test_rotor_home)

    parser_test_rotor_full = test_rotor_sub.add_parser("full", help="Test a rotor by moving it to various points", parents=[parser_control_common])
    parser_test_rotor_full.set_defaults(func=test_rotor_full)

    parser_test_radio = test_sub.add_parser("radio", help="Test a radio", parents=[parser_control_common])
    parser_test_radio.add_argument("--downlink", type=int,
                              help="Downlink frequency in herz to set the radios to")
    parser_test_radio.add_argument("--uplink", type=int,
                              help="Uplink frequency in herz to set the radios to")
    parser_test_radio.set_defaults(func=test_radio)

    # settings subcommands
    parser_settings = sub_parsers.add_parser("settings", help="View and change settings")
    settings_sub = parser_settings.add_subparsers(required=True)

    parser_settings_list = settings_sub.add_parser("list", help="List all settings")
    parser_settings_list.set_defaults(func=list_settings)

    parser_settings_modify = settings_sub.add_parser("modify", help="Modify the value of a setting")
    parser_settings_modify.add_argument("setting_key", choices=settings.load_settings_file().keys(),
                                        help="The key of the setting to modify (check with `satgs settings list`)")
    parser_settings_modify.add_argument("new_setting_value", help="The new value of the setting")
    parser_settings_modify.set_defaults(func=modify_setting)

    parser_settings_get_path = settings_sub.add_parser("path", help="Get the path in which all config files are stored")
    parser_settings_get_path.set_defaults(func=get_settings_path)

    parser_setting_clean_data = settings_sub.add_parser("clean", help="Delete all config and data files created by satgs")
    parser_setting_clean_data.set_defaults(func=clean_data)
    
    # Set default function (if no subcommand was provided) to show error message
    parser.set_defaults(func=no_args_message)

    # Parse arguments
    args = parser.parse_args()
    logging.log(logging.DEBUG, "Got args "+str(args))

    # Handle global arguments
    if args.debug:
        set_debug()

    if args.version:
        show_version(args)

    # Run subcommand associated function
    args.func(args)


from src import radio_controller, rotor_controller, tle, paths, settings, transponders
from skyfield.api import load, wgs84
from typing import List
import logging, os, datetime, time, traceback

TRACKING_UPDATE_INTERVAL = float(settings.get_setting("tracking_update_interval")) # Tracking update interval in seconds

def list_rotors() -> List[str]:
    """Return a list of all rotor config file names (excluding file extension)"""
    files = os.listdir(paths.ROTOR_CONFIG_DIRECTORY_PATH)
    files_no_extension = [file[:-5] for file in files]

    return files_no_extension

def list_radios() -> List[str]:
    """Return a list of all radio config file names (excluding file extension)"""
    files = os.listdir(paths.RADIO_CONFIG_DIRECTORY_PATH)
    files_no_extension = [file[:-5] for file in files]

    return files_no_extension

def track(NORAD_ID: str, 
          rotor_config_name: str | None = None,
          radio_config_name: str | None = None,
          rotor_usb_overwrite: str | None = None,
          rx_usb_overwrite: str | None = None,
          tx_usb_overwrite: str | None = None,
          trx_usb_overwrite: str | None = None,
          lock_up_down: bool = True,
          rotor_control_mode_overwrite: int | None = None):
    if rotor_config_name is None and radio_config_name is None:
        logging.log(logging.ERROR, "Must provide either a radio config, rotor config or both. Not none.")
        exit()

    # Initialize timescale
    timescale = load.timescale()

    # Initialize station position
    station_latitude = settings.get_setting("station_latitude")
    station_longitude = settings.get_setting("station_longitude")
    station_altitude = settings.get_setting("station_altitude")
    station_location = wgs84.latlon(float(station_latitude), float(station_longitude), float(station_altitude))

    # Try to load TLE
    satellite = tle.load_tle(NORAD_ID, timescale)
    if satellite is None:
        logging.log(logging.ERROR, "Failed to load TLE for NORAD "+NORAD_ID+". Not found in local files.")
        exit()
    else:
        if satellite.name is None:
            logging.log(logging.WARN, "Successfully loaded TLE for satellite with NORAD ID "+NORAD_ID+", but satellite name was None.")
        else:
            logging.log(logging.INFO, "Successfully loaded TLE for satellite '"+satellite.name+"'")

    # If radio is defined, prompt user to select transponder
    downlink_start = None
    uplink_start = None
    inverting = False
    if radio_config_name:
        logging.log(logging.INFO, "Please select which transponder the radio(s) should track (or 'help' for help menu):")
        transponder_UUID = transponders.user_transponder_selection(NORAD_ID)
        transponder_frequencies = transponders.get_transponder_frequencies(NORAD_ID, transponder_UUID)
        downlink_lower, downlink_upper, uplink_lower, uplink_upper, inverting = transponder_frequencies

        # Set starting frequency to middle of upper and lower downlink frequency if an upper frequency is given
        downlink_start = downlink_lower
        if downlink_upper:
            downlink_start = (downlink_lower + downlink_upper) // 2

        # The same for uplink
        uplink_start = uplink_lower
        if uplink_upper:
            uplink_start = (uplink_lower + uplink_upper) // 2

    utc_now = datetime.datetime.now(datetime.timezone.utc)
    
    # Check if pass has already begun
    pos = (satellite - station_location).at(timescale.from_datetime(utc_now))
    elevation, azimuth, _ = pos.altaz() # type: ignore
    elevation: float = elevation.degrees # type: ignore
    
    initial_elevation = 0
    pass_already_started = False
    if elevation > 0:
        pass_already_started = True
        initial_azimuth = azimuth
        initial_elevation = round(elevation)
        earliest_rise_time = datetime.datetime.now(datetime.timezone.utc)
    else:
        # Calculate time of next pass

        stop = utc_now + datetime.timedelta(hours=12)
        times, events = satellite.find_events(station_location, timescale.from_datetime(utc_now), timescale.from_datetime(stop))

        try:
            earliest_rise_index = list(events).index(0)
            earliest_rise_time = times[earliest_rise_index]
        except ValueError:
            logging.log(logging.ERROR, "No pass of satellite found within the next 12 hours.")
            exit()

        # Calculate beginnning azimuth of pass
        pos = (satellite - station_location).at(earliest_rise_time)
        _, initial_azimuth, _ = pos.altaz()
        earliest_rise_time = earliest_rise_time.utc_datetime()

        # Notify user
        logging.log(logging.INFO, f"Found next pass at {earliest_rise_time.strftime('%H:%M:%S')} UTC with an initial azimuth of {round(initial_azimuth.degrees)}°") # type: ignore
    initial_azimuth = round(initial_azimuth.degrees) # type: ignore

    # Initialize rotor
    rotor = None
    if rotor_config_name:
        rotor = rotor_controller.Rotor_Controller(rotor_config_name, rotor_usb_overwrite, rotor_control_mode_overwrite)
    
    # Initialize radio
    radio = None
    if radio_config_name:
        radio = radio_controller.Radio_Controller(radio_config_name, downlink_start, uplink_start, rx_usb_overwrite, tx_usb_overwrite, trx_usb_overwrite, inverting, lock_up_down)

    try: # From this point on, catch KeyboardInterrupt or other excpetions and make sure rot/rigctld are terminated and the sockets are closed.
        logging.log(logging.INFO, "Ready to start")
        if radio:
            # Set rig frequency to uncorrected frequency to test rig communication
            radio.update(0)
        
        # Spin rotor to pass starting angle
        if rotor:
            logging.log(logging.INFO, "Rotating to starting azimuth")
            rotor.rotate_to_blocking(initial_azimuth, initial_elevation)
            logging.log(logging.INFO, "Rotor is at start azimuth")
        
        # Wait for pass to start if pass hasn't begun yet
        if not pass_already_started:
            utc_now = datetime.datetime.now(datetime.timezone.utc)
            time_until_pass = earliest_rise_time - utc_now # type: ignore
            seconds_until_pass = time_until_pass.total_seconds()

            if seconds_until_pass > 10:
                if seconds_until_pass > 60:
                    logging.log(logging.INFO, f"Waiting for pass to start ({round(seconds_until_pass/60)} min / {earliest_rise_time.strftime('%H:%M')}z)") # type: ignore
                else:
                    logging.log(logging.INFO, f"Waiting for pass to start ({round(seconds_until_pass)}s)")
                time.sleep(seconds_until_pass-10)
                logging.log(logging.INFO, "Pass starting in 10 seconds!")
                time.sleep(10)
            elif (seconds_until_pass < 10) and (seconds_until_pass > 0):
                logging.log(logging.INFO, f"Pass starting in {round(seconds_until_pass)} seconds!")
                time.sleep(seconds_until_pass)

        # Update frequency once just before starting so first offset lock offset is calculated correctly
        if radio:
            utc_now = datetime.datetime.now(datetime.timezone.utc)
            pos = (satellite - station_location).at(timescale.from_datetime(utc_now))
            _, _, _, _, _, range_rate = pos.frame_latlon_and_rates(station_location)
            radio.update(float(range_rate.km_per_s)) # type: ignore

        peak_elevation = 0
        is_descending = 0

        while True:
            utc_now = datetime.datetime.now(datetime.timezone.utc)
            pos = (satellite - station_location).at(timescale.from_datetime(utc_now))

            # Calculate current satellite position
            elevation, azimuth, _ = pos.altaz() # type: ignore
            azimuth: int = round(azimuth.degrees) # type: ignore
            elevation: float = elevation.degrees # type: ignore

            # Update peak elevation and check if satellite elevation is descending
            if elevation > peak_elevation:
                peak_elevation = elevation
                is_descending = False
            elif elevation < peak_elevation:
                is_descending = True

            # Check if pass is done
            if is_descending:
                if elevation < 0:
                    logging.log(logging.INFO, "Pass completed!")
                    if rotor and rotor.home_on_end:
                        time.sleep(5) # Wait a bit to make sure the signal is really gone
                        logging.log(logging.INFO, "Homing rotor..")
                        rotor.rotate_to_blocking(0, 0)
                        logging.log(logging.INFO, "Done")
                    break

            # Handle rotor
            if rotor:
                # Update rotor position
                rotor.update(azimuth, round(elevation)) # type: ignore

            # Handle radios
            radio_status_msg = ""
            if radio:
                # Calculate range rate
                _, _, _, _, _, range_rate = pos.frame_latlon_and_rates(station_location)
            
                # Update frequencies
                radio.update_lock()
                radio.update(range_rate.km_per_s) # type: ignore

                # Prepare status message
                downlink_message = ""
                uplink_message = ""

                if radio.corrected_downlink:
                    current_downlink = round(radio.current_downlink_frequency/1000000, 4) # show base frequency in MHz
                    current_downlink = "{:.4f}".format(current_downlink) # make sure there's always 4 floating points (pad with zeroes)
                    doppler_shift = round(radio.downlink_correction)  # show doppler shift correction in herz
                    doppler_shift_symbol = "+" if doppler_shift >= 0 else "" # show plus if doppler shift is positive
                    downlink_message = f"D: {current_downlink}M {doppler_shift_symbol}{doppler_shift}"
                if radio.corrected_uplink:
                    current_uplink = round(radio.current_uplink_frequency/1000000, 4) # show base frequency in MHz
                    current_uplink = "{:.4f}".format(current_uplink) # make sure there's always 4 floating points (pad with zeroes)
                    doppler_shift = round(radio.uplink_correction)  # show doppler shift correction in herz
                    doppler_shift_symbol = "+" if doppler_shift >= 0 else "" # show plus if doppler shift is positive
                    uplink_message = f"U: {current_uplink}M {doppler_shift_symbol}{doppler_shift}"

                radio_status_msg = f"{downlink_message}  {uplink_message}"

            # Generate rotor status message
            az_el_msg = f"AZ: {azimuth}°  EL: {round(elevation, 1)}°   "

            # Log current status to console
            logging.log(logging.INFO, az_el_msg+radio_status_msg)

            # Wait delay
            time.sleep(TRACKING_UPDATE_INTERVAL)
    except BaseException as e:
        if isinstance(e, KeyboardInterrupt):
            logging.log(logging.INFO, "Caught keyboard interrupt, shutting down subprocesses")
        else:
            logging.log(logging.ERROR, "Caught exception, shutting down subprocesses")
            logging.log(logging.ERROR, e)
            if logging.getLogger().level == logging.DEBUG:
                logging.log(logging.DEBUG, traceback.format_exc())
    finally:
        # Close sockets and rxxctlds
        if rotor:
            rotor.close()

        if radio:
            radio.close()



from src import radio_controller, rotor_controller
import logging, time

def rotor_home(rotor_config_name: str, usb_overwrite: str | None = None, rotor_mode_overwrite: int | None = None):
    """A testing function to home a rotor to north"""

    logging.log(logging.INFO, "Initializing rotor")
    rotor = rotor_controller.Rotor_Controller(rotor_config_name, usb_overwrite, rotor_mode_overwrite)

    logging.log(logging.INFO, "Homing rotor..")
    rotor.rotate_to_blocking(0, 0) # This might output warnings with strange rotor limits
    logging.log(logging.INFO, "Done")

    rotor.close()

def rotor_test(rotor_config_name: str, usb_overwrite: str | None = None, rotor_mode_overwrite: int | None = None):
    """A test function to rotate a rotor slightly to check if it's working properly"""

    logging.log(logging.INFO, "Initializing rotor")
    rotor = rotor_controller.Rotor_Controller(rotor_config_name, usb_overwrite, rotor_mode_overwrite)

    rotor.update_current_position()
    start_azimuth: int = rotor.current_az # type: ignore
    start_elevation: int = rotor.current_el # type: ignore

    # Check if rotor is in safe range and if it is spin 10° in in each direction
    if start_elevation+10 < rotor.max_el:
        logging.log(logging.INFO, "Rotating 10° up")
        rotor.rotate_to_blocking(start_azimuth, start_elevation+10)
        rotor.rotate_to_blocking(start_azimuth, start_elevation)
    else:
        logging.log(logging.INFO, "Not rotating 10° up to not go out of possible range.")
    
    if start_elevation-10 > rotor.min_el:
        logging.log(logging.INFO, "Rotating 10° down")
        rotor.rotate_to_blocking(start_azimuth, start_elevation-10)
        rotor.rotate_to_blocking(start_azimuth, start_elevation)
    else:
        logging.log(logging.INFO, "Not rotating 10° down to not go out of possible range.")
    
    if start_azimuth+10 < rotor.max_az:
        logging.log(logging.INFO, "Rotating 10° clockwise")
        rotor.rotate_to_blocking(start_azimuth+10, start_elevation)
        rotor.rotate_to_blocking(start_azimuth, start_elevation)
    else:
        logging.log(logging.INFO, "Not rotating 10° clockwise to not go out of possible range.")

    if start_azimuth-10 > rotor.min_az:
        logging.log(logging.INFO, "Rotating 10° counterclockwise")
        rotor.rotate_to_blocking(start_azimuth-10, start_elevation)
        rotor.rotate_to_blocking(start_azimuth, start_elevation)
    else:
        logging.log(logging.INFO, "Not rotating 10° counterclockwise to not go out of possible range.")

    rotor.close()

def rotor_test_full(rotor_config_name: str, usb_overwrite: str | None = None, rotor_mode_overwrite: int | None = None):
    """A test function to rotate a rotor to a few different points to evaluate possible issues at rotor limits"""

    logging.log(logging.INFO, "Initializing rotor")
    rotor = rotor_controller.Rotor_Controller(rotor_config_name, usb_overwrite, rotor_mode_overwrite)

    max_az = rotor.max_az
    min_az = rotor.min_az
    max_el = rotor.max_el
    min_el = rotor.min_el
    el_third = max_el / 3

    TEST_POINTS = [(min_az, min_el), (max_az/2, el_third), (max_az, el_third*2), (min_az, max_el), (min_az, min_el)]

    for i, (az, el) in enumerate(TEST_POINTS):
        az = round(az)
        el = round(el)
        logging.log(logging.INFO, f"Rotating to point {i+1}/{len(TEST_POINTS)} - AZ {az} EL {el}")
        rotor.rotate_to_blocking(az, el)
    logging.log(logging.INFO, "Done")

    rotor.close()

def test_radio(radio_config_name: str, downlink_frequency: int | None, uplink_frequency: int | None, rx_usb_overwrite: str | None, tx_usb_overwrite: str | None, trx_usb_overwrite: str | None):
    """A test function to set a radio to a certain frequency (specified in herz)"""

    logging.log(logging.INFO, "Initializing radio")
    radio = radio_controller.Radio_Controller(radio_config_name, downlink_frequency, uplink_frequency, rx_usb_overwrite, tx_usb_overwrite, trx_usb_overwrite)
    
    logging.log(logging.INFO, "Updating frequency")
    radio.update(0)

    logging.log(logging.INFO, "Entering lock update loop. Press ctrl+C to exit.")
    try:
        while True:
            radio.update_lock()
            radio.update(0)
            time.sleep(1)
    except Exception:
        radio.close()



from src import util
import logging, sys, os

class CustomFormatter(logging.Formatter):
    # A custom logging formatter

    # ANSI escape codes for coloring
    RESET  = "\x1b[0m"
    RED    = "\x1b[31m"
    YELLOW = "\x1b[33m"

    FORMAT = "[%(asctime)s] (%(levelchar)s) %(message)s" # Default

    LEVEL_MAP = {
        logging.ERROR:   ("E", RED),
        logging.WARNING: ("W", YELLOW),
        logging.INFO:    ("I", ""),  # no color
        logging.DEBUG:   ("D", ""),  # no color
    }

    def format(self, record):
        levelno = record.levelno
        char, color = self.LEVEL_MAP.get(levelno, ("?", ""))

        # Inject new attributes into the record
        record.levelchar = char
        fmt = color + self.FORMAT + self.RESET
        formatter = logging.Formatter(fmt, "%H:%M:%S")
        return formatter.format(record)

def handle_uncaught(exc_type, exc_value, exc_tb):
    """A function to show uncaught exceptions by sending them to logging"""

    logging.log(logging.ERROR, "Unhandled exception: %s", exc_value)
    logging.log(logging.DEBUG, "Full traceback:", exc_info=(exc_type, exc_value, exc_tb))
    sys.exit(1)

def set_up_logging():
    """A function that sets up a basic logging system that prints to stdout"""
    # Set exceptions to be handeled by custom function
    sys.excepthook = handle_uncaught

    # Set logging level to debug if run in poetry environment
    logging_level = logging.INFO
    if util.is_poetry():
        logging_level = logging.DEBUG

    if os.name == "nt":
        from colorama import just_fix_windows_console
        just_fix_windows_console()

    # Set up logging to print to console
    logging_root = logging.getLogger()
    logging_root.setLevel(logging_level)
    logging_handler = logging.StreamHandler(sys.stdout)
    logging_handler.setLevel(logging_level)
    logging_handler.setFormatter(CustomFormatter())
    logging_root.addHandler(logging_handler)

    logging.log(logging.DEBUG, "Poetry environment detected. Setting logging level to debug.")



from src import paths, util
from typing import Dict, Tuple
import subprocess, os, json, logging, socket, time

ROTOR_CONF_EXPECTED_KEYS = set(["usb_port", "rotctl_ID", "min_az", "max_az", "min_el", "max_el", "control_type", "home_on_end"])

def parse_rotor_config(rotor_config_name: str) -> Dict[str, str | int]:
    """
    Parse a rotor config file by its file name (excluding file exension).
    Returns all values specified in the README section for the rotor config files in a dictionary.
    """
    
    file_path = os.path.join(paths.ROTOR_CONFIG_DIRECTORY_PATH, rotor_config_name+".json")
    try:
        with open(file_path, "r") as f:
            json_data = json.load(f)
    except json.JSONDecodeError as e:
        logging.log(logging.ERROR, "Failed parsing file rotor config file '"+rotor_config_name+".json'.")
        logging.log(logging.ERROR, e)
        exit()

    if type(json_data) is not dict:
        logging.log(logging.ERROR, "Failed parsing file rotor config file '"+rotor_config_name+".json'. JSON data parsed to invalid data type.")
        exit()

    if set(json_data.keys()) != ROTOR_CONF_EXPECTED_KEYS:
        logging.log(logging.ERROR, "Failed parsing file rotor config file '"+rotor_config_name+".json'. Invalid keys present in config file.")
        exit()

    return json_data

class Rotor_Controller():
    def __init__(self, rotor_config_name: str, usb_overwrite: str | None = None, rotor_mode_overwrite: int | None = None) -> None:
        """
        Initialize rotor object. Must provide the name of the rotor config file to be read, without the extension.
        Optionally, define a usb port to overwrite the one in the config.
        """

        # Parse config
        rotor_config = parse_rotor_config(rotor_config_name)

        self.usb_port = str(rotor_config["usb_port"]) if usb_overwrite is None else usb_overwrite
        self.rotctl_ID = str(rotor_config["rotctl_ID"])
        self.rotctld_port = util.get_unused_port("rotctld")
        self.min_az = int(rotor_config["min_az"])
        self.max_az = int(rotor_config["max_az"])
        self.min_el = int(rotor_config["min_el"])
        self.max_el = int(rotor_config["max_el"])
        self.control_type = int(rotor_config["control_type"]) if rotor_mode_overwrite is None else rotor_mode_overwrite
        self.home_on_end = bool(rotor_config["home_on_end"])

        # Attempt to start rotctld
        logging.log(logging.INFO, "Starting rotctld")
        set_conf_arg = "--set-conf=min_az="+str(self.min_az)+",max_az="+str(self.max_az)+",min_el="+str(self.min_el)+",max_el="+str(self.max_el)
        self.rotctld = subprocess.Popen(
            ["rotctld", "-m", str(self.rotctl_ID), "-r", str(self.usb_port), "-t", str(self.rotctld_port), set_conf_arg],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True
        )

        try:
            stdout, stderr = self.rotctld.communicate(timeout=1)
            if self.rotctld.returncode != 0:
                logging.log(logging.ERROR, "Rotctld failed with error code "+str(self.rotctld.returncode))
                logging.log(logging.ERROR, "Error: "+str(stderr))
                if stderr == "rot_open: error = IO error":
                    logging.log(logging.INFO, "Tip: Make sure you have the correct USB port selected." \
                                              "You can overwrite the USB port in the config file using -o")
                exit()
            if self.rotctld is None:
                logging.log(logging.ERROR, "Rotctld failed to start.")
        except subprocess.TimeoutExpired: # Rotctld is running
            pass

        logging.log(logging.DEBUG, "Opening socket to rotctld")
        self.sock = socket.create_connection(("localhost", int(self.rotctld_port)), timeout=3)

        self.current_az = None
        self.current_el = None

    def _send_rotctld_command(self, cmd: str):
        """
        Send a command to rotctld and return response lines (without newlines).
        """
        logging.log(logging.DEBUG, f"Sending rotctld command '{cmd}'")
        self.sock.sendall((cmd + '\n').encode('ascii'))
        response = self.sock.recv(4096).decode('ascii')
    
        # multiple lines, strip trailing newline
        return [line.strip() for line in response.splitlines()]
    
    def _apply_control_mode(self, azimuth: int, elevation: int) -> Tuple[int, int]:
        """
        Applies control mode to target azimuth/elevation to get the real position that the rotor needs to spin to.
        """

        if self.control_type == 2:
            azimuth = (round(self.max_az/2) + azimuth) % self.max_az

        return (azimuth, elevation)
    
    def rotate_to(self, azimuth: int, elevation: int):
        """
        Send rotctld command to spin rotor to a certain azimuth and elevation. Doesn't take control mode into account. 
        Automatically clamps too high/low azimuth/elevation to maximum/minimum.
        """

        # Clamp elevation value
        if elevation < self.min_el:
            elevation = self.min_el
        elif elevation > self.max_el:
            if self.max_el > 90:
                logging.log(logging.WARN, "Tried to rotate to a too high elevation on a rotor that supports elevations of more than 90°. This is likely due to a bug and will cause issues.")
            elevation = self.max_az

        # Clamp azimuth value
        if azimuth < self.min_az:
            azimuth = self.min_az
        elif azimuth > self.max_az:
            if self.max_az > 360:
                logging.log(logging.WARN, "Tried to rotate to a too high azimuth on a rotor that supports azimuths of more than 360°. This is likely due to a bug and will cause issues.")
            azimuth = self.max_az

        self._send_rotctld_command(f"P {azimuth} {elevation}")

    def update_current_position(self):
        """
        Get current rotor position and store it in the current_az and current_el variables.
        """
        resp = self._send_rotctld_command("p")
        
        self.current_az = round(float(resp[0]))
        self.current_el = round(float(resp[1]))

    def update(self, new_azimuth: int, new_elevation: int):
        """
        Update rotor movement with new target elevation and azimuth values.
        """
        
        # Read azimuth and elevation
        self.update_current_position()

        # Apply alternate control style if option is set
        new_azimuth, new_elevation = self._apply_control_mode(new_azimuth, new_elevation)

        self.rotate_to(new_azimuth, new_elevation)

    def rotate_to_blocking(self, azimuth: int, elevation: int, tolerance: int = 2):
        """
        Spins the rotor to a certain position and blocks (sleeps) until it has reached the target position.
        Requires target azimuth, elevation, and inaccuracy tolerance in degrees.
        """

        # If alternate control type is used, make sure the correct target azimuth is being checked for
        target_azimuth, target_elevation = self._apply_control_mode(azimuth, elevation)

        # Check if target is in supported rotor range (TODO: in non-blocking rotation this is done silently. Change that?)
        if target_azimuth < self.min_az:
            if abs(target_azimuth - self.min_az) > tolerance:
                logging.log(logging.WARN, f"Target azimuth {target_azimuth} is too far from minimum possible azimuth to be in tolerance. Setting target to closest azimuth possible.")
                target_azimuth = self.min_az
        elif target_azimuth > self.max_az:
            if abs(target_azimuth - self.max_az) > tolerance:
                logging.log(logging.WARN, f"Target azimuth {target_azimuth} is too far from maximum possible azimuth to be in tolerance. Setting target to closest azimuth possible.")
                target_azimuth = self.max_az

        if target_elevation < self.min_el:
            if abs(target_elevation - self.min_el) > tolerance:
                logging.log(logging.WARN, f"Target elevation {target_elevation} is too far from minimum possible elevation to be in tolerance. Setting target to closest elevation possible.")
                target_elevation = self.min_el
        elif target_elevation > self.max_el:
            if abs(target_elevation - self.max_el) > tolerance:
                logging.log(logging.WARN, f"Target elevation {target_elevation} is too far from maximum possible elevation to be in tolerance. Setting target to closest elevation possible.")
                target_elevation = self.min_el

        # Rotate to target location
        self.update(azimuth, elevation)

        # Check if current rotation is within tolerated range of target angles
        while (abs(target_azimuth-self.current_az) > 2) or (abs(target_elevation-self.current_el) > 2): # type: ignore
            self.update_current_position()
            logging.log(logging.DEBUG, f"Rotating to target - AZ {self.current_az} -> {target_azimuth}  EL {self.current_el} -> {target_elevation}")
            time.sleep(0.5)

    def close(self):
        """Close socket and terminate rotctl instance"""
        logging.log(logging.DEBUG, "Closing rotor controller")

        self.sock.close()
        self.rotctld.terminate()

